# This file is part of pydpc.
#
# Copyright 2016 Christoph Wehmeyer
#
# pydpc is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as _np
import matplotlib.pyplot as _plt
from . import core as _core

__all__ = ['Cluster']

class Distances(object):
    def __init__(self, points):
        self.points = points
        self.npoints = self.points.shape[0]
        self.distances = _core.get_distances(self.points)
        self.max_distance = self.distances.max()

class Density(Distances):
    def __init__(self, points, fraction):
        super(Density, self).__init__(points)
        self.fraction = fraction
        self.kernel_size = _core.get_kernel_size(self.distances, self.fraction)
        self.density = _core.get_density(self.distances, self.kernel_size)

class Graph(Density):
    def __init__(self, points, fraction, autoplot):
        super(Graph, self).__init__(points, fraction)
        self.autoplot = autoplot
        self.order = _np.ascontiguousarray(_np.argsort(self.density).astype(_np.intc)[::-1])
        self.delta, self.neighbour = _core.get_delta_and_neighbour(
            self.order, self.distances, self.max_distance)
        if self.autoplot:
            self.draw_decision_graph()
    def draw_decision_graph(self, min_density=None, min_delta=None):
        fig, ax = _plt.subplots(figsize=(8, 4.5))
        ax.scatter(self.density, self.delta, s=40)
        if min_density is not None and min_delta is not None:
            ax.plot(
                [min_density, self.density.max()], [min_delta, min_delta], linewidth=2, color="red")
            ax.plot(
                [min_density, min_density], [min_delta, self.delta.max()], linewidth=2, color="red")
        ax.set_xlabel(r"density", fontsize=20)
        ax.set_ylabel(r"delta / a.u.", fontsize=20)
        ax.tick_params(labelsize=15)
        return fig, ax

class Cluster(Graph):
    def __init__(self, points, fraction=0.02, autoplot=True):
        super(Cluster, self).__init__(points, fraction, autoplot)
    def assign(self, min_density, min_delta, border_only=False):
        self.min_density = min_density
        self.min_delta = min_delta
        self.border_only = border_only
        if self.autoplot:
            self.draw_decision_graph(self.min_density, self.min_delta)
        self._get_cluster_indices()
        self.membership = _core.get_membership(self.clusters, self.order, self.neighbour)
        self.border_density, self.border_member = _core.get_border(
            self.kernel_size, self.distances, self.density, self.membership, self.nclusters)
        self.halo_idx, self.core_idx = _core.get_halo(
            self.density, self.membership,
            self.border_density, self.border_member.astype(_np.intc), border_only=border_only)
    def _get_cluster_indices(self):
        self.clusters = _np.intersect1d(
            _np.where(self.density > self.min_density)[0],
            _np.where(self.delta > self.min_delta)[0], assume_unique=True).astype(_np.intc)
        self.nclusters = self.clusters.shape[0]



# This file is part of pydpc.
#
# Copyright 2016 Christoph Wehmeyer
#
# pydpc is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pydpc._reference import Cluster as Ref
from pydpc import Cluster
import numpy as np
from numpy.testing import assert_array_equal, assert_almost_equal

class TestFourGaussians2D(object):
    @classmethod
    def setup_class(cls):
        # data generation
        cls.npoints = 1000
        cls.mux = 1.8
        cls.muy = 1.8
        cls.fraction = 0.02
        cls.points = np.zeros(shape=(cls.npoints, 2), dtype=np.float64)
        cls.points[:, 0] = np.random.randn(cls.npoints) + \
            cls.mux * (-1)**np.random.randint(0, high=2, size=cls.npoints)
        cls.points[:, 1] = np.random.randn(cls.npoints) + \
            cls.muy * (-1)**np.random.randint(0, high=2, size=cls.npoints)
        # cluster initialisation
        cls.ref = Ref(cls.fraction, autoplot=False)
        cls.ref.load(cls.points)
        cls.ref.assign(20, 1.5)
        cls.dpc = Cluster(cls.points, cls.fraction, autoplot=False)
        cls.dpc.assign(20, 1.5)
    @classmethod
    def teardown_class(cls):
        pass
    def setup(self):
        pass
    def teardown(self):
        pass
    def test_distances(self):
        assert_almost_equal(self.dpc.distances, self.ref.distances, decimal=10)
    def test_distances(self):
        assert_almost_equal(self.dpc.kernel_size, self.ref.kernel_size, decimal=10)
    def test_density(self):
        assert_almost_equal(self.dpc.density, self.ref.density, decimal=10)
    def test_order(self):
        assert_array_equal(self.dpc.order, self.ref.order)
    def test_delta(self):
        assert_almost_equal(self.dpc.delta, self.ref.delta, decimal=10)
    def test_neighbour(self):
        assert_array_equal(self.dpc.neighbour, self.ref.neighbour)
    def test_clusters(self):
        assert_array_equal(self.dpc.clusters, self.ref.clusters)
    def test_membership(self):
        assert_array_equal(self.dpc.membership, self.ref.membership)
    def test_border_density(self):
        assert_almost_equal(self.dpc.border_density, self.ref.border_density, decimal=10)
    def test_border_member(self):
        assert_array_equal(self.dpc.border_member, self.ref.border_member)
    def test_halo_idx(self):
        assert_array_equal(self.dpc.halo_idx, self.ref.halo_idx)
    def test_core_idx(self):
        assert_array_equal(self.dpc.core_idx, self.ref.core_idx)



# This file is part of pydpc.
#
# Copyright 2016 Christoph Wehmeyer
#
# pydpc is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as _np
cimport numpy as _np

cdef extern from "_core.h":
    void _get_distances(double *points, int npoints, int ndim, double *distances)
    double _get_kernel_size(double *distances, int npoints, double fraction)
    void _get_density(double kernel_size, double *distances, int npoints, double *density)
    void _get_delta_and_neighbour(
        double max_distance, double *distances, int *order,
        int npoints, double *delta, int *neighbour)
    void _get_membership(
        int *clusters, int nclusters, int *order, int *neighbour, int npoints, int *membership)
    void _get_border(
        double kernel_size, double *distances, double *density, int *membership, int npoints,
        int *border_member, double *border_density)
    void _get_halo(
        int border_only, double *border_density,
        double *density, int *membership, int *border_member, int npoints, int *halo)

def get_distances(_np.ndarray[double, ndim=2, mode="c"] points not None):
    npoints = points.shape[0]
    ndim = points.shape[1]
    distances = _np.zeros(shape=(npoints, npoints), dtype=_np.float64)
    _get_distances(
        <double*> _np.PyArray_DATA(points),
        npoints, ndim,
        <double*> _np.PyArray_DATA(distances))
    return distances

def get_kernel_size(_np.ndarray[double, ndim=2, mode="c"] distances not None, fraction):
    return _get_kernel_size(<double*> _np.PyArray_DATA(distances), distances.shape[0], fraction)

def get_density(_np.ndarray[double, ndim=2, mode="c"] distances not None, kernel_size):
    npoints = distances.shape[0]
    density = _np.zeros(shape=(npoints,), dtype=_np.float64)
    _get_density(
        kernel_size,
        <double*> _np.PyArray_DATA(distances),
        npoints,
        <double*> _np.PyArray_DATA(density))
    return density

def get_delta_and_neighbour(
    _np.ndarray[int, ndim=1, mode="c"] order not None,
    _np.ndarray[double, ndim=2, mode="c"] distances not None,
    max_distance):
    npoints = distances.shape[0]
    delta = _np.zeros(shape=(npoints,), dtype=_np.float64)
    neighbour = _np.zeros(shape=(npoints,), dtype=_np.intc)
    _get_delta_and_neighbour(
        max_distance,
        <double*> _np.PyArray_DATA(distances),
        <int*> _np.PyArray_DATA(order),
        npoints,
        <double*> _np.PyArray_DATA(delta),
        <int*> _np.PyArray_DATA(neighbour))
    return delta, neighbour

def get_membership(
    _np.ndarray[int, ndim=1, mode="c"] clusters not None,
    _np.ndarray[int, ndim=1, mode="c"] order not None,
    _np.ndarray[int, ndim=1, mode="c"] neighbour not None):
    npoints = order.shape[0]
    membership = _np.zeros(shape=(npoints,), dtype=_np.intc)
    _get_membership(
        <int*> _np.PyArray_DATA(clusters),
        clusters.shape[0],
        <int*> _np.PyArray_DATA(order),
        <int*> _np.PyArray_DATA(neighbour),
        npoints,
        <int*> _np.PyArray_DATA(membership))
    return membership

def get_border(
    kernel_size,
    _np.ndarray[double, ndim=2, mode="c"] distances not None,
    _np.ndarray[double, ndim=1, mode="c"] density not None,
    _np.ndarray[int, ndim=1, mode="c"] membership not None,
    nclusters):
    npoints = distances.shape[0]
    border_density = _np.zeros(shape=(nclusters,), dtype=_np.float64)
    border_member = _np.zeros(shape=(npoints,), dtype=_np.intc)
    _get_border(
        kernel_size,
        <double*> _np.PyArray_DATA(distances),
        <double*> _np.PyArray_DATA(density),
        <int*> _np.PyArray_DATA(membership),
        npoints,
        <int*> _np.PyArray_DATA(border_member),
        <double*> _np.PyArray_DATA(border_density))
    return border_density, border_member.astype(_np.bool)

def get_halo(
    _np.ndarray[double, ndim=1, mode="c"] density not None,
    _np.ndarray[int, ndim=1, mode="c"] membership not None,
    _np.ndarray[double, ndim=1, mode="c"] border_density not None,
    _np.ndarray[int, ndim=1, mode="c"] border_member not None,
    border_only=False):
    halo = membership.copy()
    flag = 0
    if border_only:
        flag = 1
    _get_halo(
        flag,
        <double*> _np.PyArray_DATA(border_density),
        <double*> _np.PyArray_DATA(density),
        <int*> _np.PyArray_DATA(membership),
        <int*> _np.PyArray_DATA(border_member),
        density.shape[0],
        <int*> _np.PyArray_DATA(halo))
    halo_idx = _np.where(halo == -1)[0].astype(_np.intc)
    core_idx = _np.where(halo != -1)[0].astype(_np.intc)
    return halo_idx, core_idx



# This file is part of pydpc.
#
# Copyright 2016 Christoph Wehmeyer
#
# pydpc is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""This module provides a previously tested development version. It is rather
slow and only used for consistency checks"""

import numpy as _np
import matplotlib.pyplot as _plt

class Cluster(object):
    def __init__(self, fraction=0.02, autoplot=True):
        self.fraction = fraction
        self.autoplot = autoplot
    def load(self, points):
        self.points = points
        self.npoints = self.points.shape[0]
        self._get_distances()
        self._get_kernel_size()
        self._get_density()
        self._get_delta_and_neighbour()
        if self.autoplot:
            self.draw_decision_graph()
    def draw_decision_graph(self, min_density=None, min_delta=None):
        fig, ax = _plt.subplots(figsize=(8, 4.5))
        ax.scatter(self.density, self.delta, s=40)
        if min_density is not None and min_delta is not None:
            ax.plot(
                [min_density, self.density.max()], [min_delta, min_delta], linewidth=2, color="red")
            ax.plot(
                [min_density, min_density], [min_delta, self.delta.max()], linewidth=2, color="red")
        ax.set_xlabel(r"density", fontsize=20)
        ax.set_ylabel(r"delta / a.u.", fontsize=20)
        ax.tick_params(labelsize=15)
    def assign(self, min_density, min_delta, border_only=False):
        self.min_density = min_density
        self.min_delta = min_delta
        self.border_only = border_only
        if self.autoplot:
            self.draw_decision_graph(self.min_density, self.min_delta)
        self._get_cluster_indices()
        self._get_membership()
        self._get_halo()
    def _get_distances(self):
        self.distances = _np.zeros(shape=(self.npoints, self.npoints), dtype=_np.float64)
        for i in range(self.npoints - 1):
            for j in range(i + 1, self.npoints):
                self.distances[i, j] = _np.linalg.norm(self.points[i, :] - self.points[j, :])
                self.distances[j, i] = self.distances[i, j]
    def _get_kernel_size(self):
        arr = self.distances[0, 1:]
        for i in range(1, self.npoints - 1):
            arr = _np.hstack((arr, self.distances[i, i + 1:]))
        arr = _np.sort(arr)
        imax = int(_np.floor(0.5 + self.fraction * arr.shape[0]))
        self.kernel_size = arr[imax]
    def _get_density(self):
        self.density = _np.zeros(shape=(self.npoints,), dtype=_np.float64)
        for i in range(self.npoints - 1):
            for j in range(i + 1, self.npoints):
                rho = _np.exp(-(self.distances[i, j] / self.kernel_size)**2)
                self.density[i] += rho
                self.density[j] += rho
    def _get_delta_and_neighbour(self):
        self.order = _np.argsort(self.density)[::-1]
        max_distance = self.distances.max()
        self.delta = _np.zeros(shape=self.order.shape, dtype=_np.float64)
        self.delta[self.order[0]] = -1.0
        self.delta[self.order[1:]] = max_distance
        self.neighbour = _np.empty_like(self.order)
        self.neighbour[:] = -1
        for i in range(1, self.npoints):
            for j in range(i):
                if self.distances[self.order[i], self.order[j]] < self.delta[self.order[i]]:
                    self.delta[self.order[i]] = self.distances[self.order[i], self.order[j]]
                    self.neighbour[self.order[i]] = self.order[j]
        self.delta[self.order[0]] = self.delta.max()
    def _get_cluster_indices(self):
        self.clusters = _np.intersect1d(
            _np.where(self.density > self.min_density)[0],
            _np.where(self.delta > self.min_delta)[0], assume_unique=True)
        self.ncl = self.clusters.shape[0]
    def _get_membership(self):
        self.membership = -1 * _np.ones(shape=self.order.shape, dtype=_np.intc)
        for i in range(self.ncl):
            self.membership[self.clusters[i]] = i
        for i in range(self.npoints):
            if self.membership[self.order[i]] == -1:
                self.membership[self.order[i]] = self.membership[self.neighbour[self.order[i]]]
    def _get_halo(self):
        self.halo = self.membership.copy()
        self.border_density = _np.zeros(shape=(self.ncl,), dtype=_np.float64)
        self.border_member = _np.zeros(shape=self.membership.shape, dtype=_np.bool)
        for i in range(self.npoints - 1):
            for j in range(i + 1, self.npoints):
                if (self.membership[i] != self.membership[j]) and (self.distances[i, j] < self.kernel_size):
                    average_density = 0.5 * (self.density[i] + self.density[j])
                    if self.border_density[self.membership[i]] < average_density:
                        self.border_density[self.membership[i]] = average_density
                    if self.border_density[self.membership[j]] < average_density:
                        self.border_density[self.membership[j]] = average_density
                    self.border_member[i] = True
                    self.border_member[j] = True
        if self.border_only:
            for i in range(self.npoints):
                if (self.density[i] < self.border_density[self.membership[i]]) and self.border_member[i]:
                    self.halo[i] = -1
        else:
            for i in range(self.npoints):
                if (self.density[i] < self.border_density[self.membership[i]]):
                    self.halo[i] = -1
        self.halo_idx = _np.where(self.halo == -1)[0]
        self.core_idx = _np.where(self.halo != -1)[0]



#!/usr/bin/env python

# This file is part of pydpc.
#
# Copyright 2016 Christoph Wehmeyer
#
# pydpc is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, Extension
import versioneer

def extensions():
    from numpy import get_include
    from Cython.Build import cythonize
    ext_core = Extension(
        "pydpc.core",
        sources=["ext/core.pyx", "ext/_core.c"],
        include_dirs=[get_include()],
        extra_compile_args=["-O3", "-std=c99"])
    exts = [ext_core]
    return cythonize(exts)

class lazy_cythonize(list):
    """evaluates extension list lazyly.
    pattern taken from http://tinyurl.com/qb8478q"""
    def __init__(self, callback):
        self._list, self.callback = None, callback
    def c_list(self):
        if self._list is None: self._list = self.callback()
        return self._list
    def __iter__(self):
        for e in self.c_list(): yield e
    def __getitem__(self, ii): return self.c_list()[ii]
    def __len__(self): return len(self.c_list())

def long_description():
    ld = "Clustering by fast search and find of density peaks, designed by Alex Rodriguez"
    ld += " and Alessandro Laio, is a density-based clustering algorithm. The pydpc package"
    ld += " aims to make this algorithm available for Python users."
    return ld

setup(
    cmdclass=versioneer.get_cmdclass(),
    ext_modules=lazy_cythonize(extensions),
    name='pydpc',
    version=versioneer.get_version(),
    description='A Python interface to Density Peak Clustering',
    long_description=long_description(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: C',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics'],
    keywords=[
        'cluster',
        'density'],
    url='https://github.com/cwehmeyer/pydpc',
    author='Christoph Wehmeyer',
    author_email='christoph.wehmeyer@fu-berlin.de',
    license='LGPLv3+',
    setup_requires=[
        'numpy>=1.7',
        'cython>=0.20',
        'setuptools>=0.6'],
    tests_require=[
        'numpy>=1.7',
        'nose>=1.3'],
    install_requires=[
        'numpy>=1.7',
        'matplotlib'],
    packages=['pydpc'],
    test_suite='nose.collector',
    scripts=[]
)

